print("import: 'transformers'")
import transformers

print("import: 'transformers.benchmark'")
import transformers.benchmark

print("import: 'transformers.commands'")
import transformers.commands

print("import: 'transformers.data'")
import transformers.data

print("import: 'transformers.data.datasets'")
import transformers.data.datasets

print("import: 'transformers.data.metrics'")
import transformers.data.metrics

print("import: 'transformers.data.processors'")
import transformers.data.processors

print("import: 'transformers.models'")
import transformers.models

print("import: 'transformers.pipelines'")
import transformers.pipelines

print("import: 'transformers.sagemaker'")
import transformers.sagemaker

print("import: 'transformers.utils'")
import transformers.utils

